
# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2013, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

#  require 'lib/tomte/crypto/pki'
#
#  => true
#
#  i = Tomte::PKI::Identity.new( :organization=>"widmore industries", :name => "Henry Gale", :email=>"henry@widmore.com" )
#
#  => #<Tomte::PKI::Identity organization="widmore industries", name="Henry Gale", email="henry@widmore.com">
#
#  p = Tomte::PKI::CAEntity.new(i)
#
#  => Tomte::PKI::CAEntity: identity=#<Tomte::PKI::Identity organization="widmore industries", name="Henry Gale", email="henry@widmore.com">
#
#  p.init_all
#
#  => #<OpenSSL::X509::Certificate subject=/O=widmore industries/CN=Henry Gale/emailAddress=henry@widmore.com,
#      issuer=/O=widmore industries/CN=Henry Gale/emailAddress=henry@widmore.com, serial=239320917995777674014667630812724471430,
#      not_before=2010-03-12 18:46:08 UTC, not_after=2011-03-12 18:46:08 UTC>

#  p.ca
#
#  => Tomte::PKI::CA: identity=#<Tomte::PKI::Identity organization="widmore industries", name="Henry Gale", email="henry@widmore.com">
#
#  p.cert.to_s
#
#  -----BEGIN CERTIFICATE-----
#  MIICuTCCAiKgAwIBAgIRALQLiEAitEfXrLtV9cNEJoYwDQYJKoZIhvcNAQEFBQAw
#  [...]
#  jY7r+iGFI2QWdd1frbuVi48KZLkyH+ph0JltKxA=
#  -----END CERTIFICATE-----
#
#  puts p.key
#
#  -----BEGIN RSA PRIVATE KEY-----
#  MIICXAIBAAKBgQDRaQYU3yFhNVpf5DaFD/j66kC+QQyR6+x7ROAdbu4XkDYsA29x
#  [...]
#  XKie1swqpWp6IsJ9WGuaQfsoEF4EjoPNTOE/hNCCg+s=
#  -----END RSA PRIVATE KEY-----


require 'openssl'
require 'uuidtools'
require 'ostruct'

module Tomte
  #
  # Implementation of a PKI for Tomte
  #
  class PKI

    DEFAULT_KEY_SIZE = 1024
    DEFAULT_DURATION = 3600*24*365

    DEFAULT_SYMMETRIC_CIPHER = 'aes-256-cbc'

    CA_CERT_BASE_EXTENSIONS = [
      ["basicConstraints","CA:TRUE",true],
      ["keyUsage","keyCertSign, cRLSign",true],
      ["subjectKeyIdentifier","hash",false],
      ["authorityKeyIdentifier","keyid:always",false]
    ].freeze

    CLIENT_CERT_BASE_EXTENSIONS = [
      ["keyUsage","Non Repudiation, Digital Signature, Key Encipherment",true],
      ["basicConstraints","CA:FALSE",true],
      ["subjectKeyIdentifier","hash",false],
      ["authorityKeyIdentifier","keyid:always",false],
      ["extendedKeyUsage","clientAuth, emailProtection, codeSigning",false]
    ].freeze

    #
    # used to transport identities and converti them to X.509 dns
    #
    #  i = Tomte::PKI::Identity.new(
    #     :organization=>"windmore industries",
    #     :name => "Henry Game",
    #     :email=> "gale@widmore.com" )
    #
    #  => #<Tomte::PKI::Identity organization="windmore industries", name="Henry Game", email="gale@widmore.com">
    #
    #
    class Identity < OpenStruct

      PARTS = { :country => 'C',
                :state => 'ST',
                :location => 'L',
                :organization => 'O',
                :unit => 'OU',
                :name => 'CN',
                :email => 'emailAddress'}.freeze

      attr_accessor :id

      #
      # returns an x.509 DN
      #
      #  i.to_x509_dn
      #
      #  => /O=windmore industries/CN=Henry Game/emailAddress=gale@widmore.com
      #
      def to_x509_dn
        dn = ''
        PARTS.each do |k, n|
          dn += "/#{n}=#{self.send(k)}" unless self.send(k).nil?
        end
        OpenSSL::X509::Name.parse(dn)
      end

      def from_x509_dn(dn)
        dn.split('/').each do |p|
          m = p.match(/^(.*)=(.*)$/)
          if m
            k = PARTS.key(m[1])
            send "#{k}=", m[2]
          end
        end
      end

      def self.from_x509_dn(dn)
        n = new
        n.from_x509_dn(dn)
        n
      end
    end



    #
    # base class for PKI objects
    #
    class Base


      # the X.509 certificate OpenSSL::X509::Certificate for this Entity
      attr_accessor :cert

      # the private key OpenSSL::PKey::RSA
      attr_accessor :key

      # tunable parameters for the PKI
      attr_accessor :key_size, :certificale_lifetime, :certificate_extensions


      # Initializes the PKI from an Identity
      #
      #  id = Tomte::PKI::Identity.new do |i|
      #    i.name = "Henry Gale"
      #    i.email = "henry@widmore.com"
      #  end
      #
      #  pki = Tomte::PKI::Entity.new(identity)
      #
      def initialize(identity, options = {})
        @identity = identity
        @key_size = options[:key_size] || DEFAULT_KEY_SIZE
        @certificate_lifetime =  options[:certificate_lifetime] || DEFAULT_DURATION
        @certificate_extensions =  options[:certificate_extensions] || CLIENT_CERT_BASE_EXTENSIONS
        @cipher = options[:cipher] || DEFAULT_SYMMETRIC_CIPHER
      end

      #
      # true if this PKI object is a CA
      #
      def ca?
        true if (@ca && @ca.key && @ca.cert)
      end

      #
      # true if this PKI entity is ready to work
      #
      def ready?
        true if (@key && @cert)
      end

      #
      # generates the local private key
      #
      def init_local_key(size = DEFAULT_KEY_SIZE)
        @key = generate_key(size)
      end

      #
      # does all the initialization
      #
      def init_local(keysize = DEFAULT_KEY_SIZE, options = {})
        init_local_key keysize
        init_local_certificate options
      end

      #
      # load personal key from file
      #
      def load_key(file, password)
        @key = load_a_key(file, password)
      end

      #
      # load personal certificate from file
      #
      def load_cert(file)
        cert = load_a_cert(file)
        @cert = cert if(cert.subject == @identity.to_x509_dn)
      end

      #
      # load both key/cert from file
      #
      def load_bundle(file)
        @key = load_key(file)
        @cert = load_cert(file)
      end

      #
      # writes local key to file
      #
      def write_key(file)
        write_a_key(@key, file)
      end

      #
      # writes my cert to file
      #
      def write_cert(file)
        write_a_cert(@cert, file)
      end

      #
      # writes both CA key/cert to file
      #
      def write_bundle(file)
        write_a_bundle(@key, @cert, file)
      end

      #######################################################################

      #
      # returns a random UUID
      #
      def random_uuid
        UUIDTools::UUID.random_create
      end

      #
      # write down a key/certificate bundle
      #
      def write_a_bundle(key, cert, file)
        File.open(file, "w") { |f| f.write key.to_pem; f.write cert.to_pem  }
      end

      #
      # write a certificate to file
      #
      def write_a_cert(cert, file)
        File.open(file, "w") { |f| f.write cert.to_pem }
      end

      #
      # write a key to file
      #
      def write_a_key(key, file)
        File.open(file, "w") { |f| f.write key.to_pem } if key
      end


      #
      # loads a key from file
      #
      def load_a_key(file, password = nil)
        OpenSSL::PKey::RSA.new(File.read(file), password)
      end

      #
      # loads a certificate from file
      #
      def load_a_cert(file)
        OpenSSL::X509::Certificate.new(File.read(file))
      end

      #
      # generate a private key
      #
      def generate_key(size = DEFAULT_KEY_SIZE)
        OpenSSL::PKey::RSA.generate(size)
      end

      #
      # generate a certificate request
      #
      def generate_certificate(identity, key)
        cert = OpenSSL::X509::Certificate.new
        cert.version = 2
        cert.subject = identity.to_x509_dn
        cert.public_key = key.public_key
        cert
      end

      ## BLOCK CIPHERS ##########################################################

      #
      # generic block cipher
      #
      def block_cipher(key, text, direction = :encrypt, salt=nil)
        cipher = OpenSSL::Cipher::Cipher.new(@cipher).send(direction)
        cipher.pkcs5_keyivgen(key, salt)
        cipher.update(plaintext) + cipher.final
      end

      #
      # block encrypt
      #
      def block_decrypt(key, ciphertext, salt=nil)
        block_cipher(key, ciphertext, :decrypt, salt)
      end

      #
      # block decrypt
      #
      def block_encrypt(key, plaintext, salt=nil)
        block_cipher(key, ciphertext, :encrypt, salt)
      end

      ###############################################################################

      def inspect
        "#{self.class.name}: identity=#{@identity.inspect}"
      end

      class << self
        #
        # given a bundle, generates an Entity for that certificate
        #
        def create_from_bundle(file)
          i = Identity.new()
          p = new(i)
          p.cert = p.load_a_cert(file)
          i.from_x509_dn(p.cert.subject.to_s)
          p.key = p.load_a_key(file)
          p
        end
      end

    end

    class Entity < Base

      #
      # generate a certificate for my identity, optionally signs it, if the PKI object is also a CA
      #
      def init_local_certificate(options = {})
        options[:key_size] ||= @key_size
        gen_my_key(options[:key_size]) unless @key
        @cert = generate_certificate(@identity, @key)
      end

    end



    #
    # Ca class
    #
    class CA < Base

      #
      # initiazies a CA, wants the CA
      #
      def initialize(identity, options = {})
        super
        @ca_certificate_extensions =  options[:ca_certificate_extensions] || CA_CERT_BASE_EXTENSIONS
      end

      #
      # sign the given certificate request
      #
      def sign_certificate(req, options={})

        raise ArgumentError, "not initialized" unless ready?

        not_before = options[:not_before] || Time.now
        not_after = options[:not_after] || (Time.now + (options[:certificate_lifetime] || @certificate_lifetime))
        serial = options[:serial] || random_uuid.to_i
        extensions = options[:certificate_extensions] || @certificate_extensions
        do_sign_certificate(req, serial, not_before, not_after, @cert, @key, extensions)
      end

      #
      # generates a CA certificate, a CA key, if necessary
      #
      def init_local_certificate(options = {})
        options[:key_size] ||= DEFAULT_KEY_SIZE
        initialize_local_key(options[:key_size]) unless @key

        not_before = options[:not_before] || Time.now
        not_after = options[:not_after] || (Time.now + (options[:certificate_lifetime] || @certificate_lifetime))
        serial = options[:serial] || random_uuid.to_i
        extensions = options[:ca_certificate_extensions] || @ca_certificate_extensions
        @cert = do_issue_certificate(@identity, @key, serial, not_before, not_after, extensions)
      end



      #######################################################################

      #
      # sign a certificate
      #
      def do_sign_certificate(cert, serial, not_before, not_after, issuer, issuer_key, extensions = [],
          digest = OpenSSL::Digest::SHA1.new)

        cert.serial = serial
        cert.issuer = issuer.subject
        cert.not_before = not_before
        cert.not_after = not_after
        ef = OpenSSL::X509::ExtensionFactory.new
        ef.subject_certificate = cert
        ef.issuer_certificate = issuer
        extensions.each{|oid, value, critical|
          cert.add_extension(ef.create_extension(oid, value, critical))
        }
        cert.sign(issuer_key, digest)
        cert
      end

      #
      # issue a certificate for the given identity (ie: generate and sign)
      # if issuer data is omitted, certificate will be self-signed
      #
      def do_issue_certificate(identity,
                            key,
                            serial,
                            not_before,
                            not_after,
                            extensions = [],
                            issuer = nil,
                            issuer_key = nil,
                            digest = OpenSSL::Digest::SHA1.new)

        cert = generate_certificate(identity, key)
        issuer = cert unless issuer
        issuer_key = key unless issuer_key
        do_sign_certificate(cert, serial, not_before, not_after, issuer, issuer_key, extensions, digest)
      end

    end

    #
    # An Entity who is both a Subject and a CA
    #
    class CAEntity < Entity

      attr_accessor :ca

      class << self
        #
        # given two bundles (key+cert) files, create a new CAEntity with the subject specified on the certificate
        #
        def create_from_bundles(id_file, ca_file)
          entity = Entity.create_from_bundle(id_file)
          ca = CA.create_from_bundle(ca_file)
          entity.ca = ca
          entity
        end

      end

      #
      # initializes a new CAEntity and optionally  initializes all the keys
      #
      # [:init_all] initialize all the keys
      #
      def initialize(identity, options = {})
        super identity
        @ca = CA.new(options[:ca_identity] || identity)
        init_all if options[:init_all]
      end

      #
      # initializes and enroll the embedded CA
      #
      def init_ca(keysize = DEFAULT_KEY_SIZE, options = {})
        @ca.init_local keyzise, options
      end

      #
      # initializes and enroll local keypair and the embedded CA
      #
      def init_all(keysize = DEFAULT_KEY_SIZE, options = {})
        ca.init_local keysize, options
        init_local keysize, options
      end

      #
      # generate a certificate for my identity, and signs it with the local CA
      #
      def init_local_certificate(options = {})
        super
        @cert = ca.sign_certificate(@cert, options)
      end

    end

  end
end

