
# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2013, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

require 'forwardable'
require 'optparse'
require 'compat/hash'
require 'uuidtools'
require 'socket'
require 'yaml'
require 'pp'
require 'erb'

module Tomte

  class Configuration

    SEPARATOR = ('-'*120).yellow

    extend Forwardable

    # delegate, for now, we will see if we need a semaphore
    def_delegators :@config, :[], :[]=

    # where we got the core configuration
    attr_reader :config_file

    public
    ######

    DEFAULT_FILE_QUEUE = :main

    DEFAULT_ENVIRONMENT_FILE = File.join('/etc', 'tomte', 'default_environment')

    DEFAULT_CONFIG_FILES = [
        {:path =>  File.join(Tomte::LIBRARY, '..', 'config'), :file =>  'tomte.yml', :queue => :before },
        {:path =>  :system,                                   :file =>  'tomte.yml'},
        {:path =>  File.join(Dir.getwd, 'config'),            :file =>  'tomte.yml'}
    ]


    DEFAULT_CORE_CONFIGURATION = {
      :configured => false,
      :environment => 'production',
      :common_environment => '_common',
      :log_level => nil,                  # nil means use the logger profile
      :log_profile => :default,
      :debug => false,
      :interactive => false,
      :application => 'tomte-app',
      :system_configuration_path =>  File.expand_path('tomte', '/etc' )
    }


    def initialize(args = {})

      @semaphore = Mutex.new
      @additional_options_blocks = {}
      @sections = []
      @overrides = {}

      @config = default_configuration

      @options = {}
      [:core, :local].each { |s| add_section s }

      @root = args[:root]

      add_config_files(DEFAULT_CONFIG_FILES)
      default_environment
    end

    #
    # used for safe change of configuration
    #
    def change_configuration(&blk)
      @semaphore.synchronize do
        blk.call(@config) if blk
      end
    end

    #
    # used to apply a piece of configuration
    #
    def apply_configuration(ary)
      change_configuration do |c|
        c.recursive_merge! ary
      end
    end
    alias :merge :apply_configuration


    #
    # returns a specific config file queue
    #
    def config_file_queue(queue = DEFAULT_FILE_QUEUE)
      config_file_queues[queue] ||= []
    end

    #
    # returns all the file queues
    #
    def config_file_queues
      @config[:core][:config_file_queues] ||= {}
    end

    #
    # adds all the files matching glob in path
    #
    def add_config_files_matching(glob, options, queue = DEFAULT_FILE_QUEUE)
      glob = expand_path_spec(glob, options[:path])
      files = Dir.glob(glob).sort.map { |f| options.merge(:file => f) }
      add_config_files(files, queue)
    end

    #
    # add one or more config files to a file queue
    #
    def add_config_files(fileinfo, queue = DEFAULT_FILE_QUEUE)
      if fileinfo.is_a? Array
        fileinfo.each { |f | add_config_file(f , queue) }
      else
        add_config_file(fileinfo, queue)
      end
    end

    #
    # adds one file to a file queue
    #
    def add_config_file(fileinfo, queue = DEFAULT_FILE_QUEUE)
      case fileinfo
        when Hash
          queue = (fileinfo.delete(:queue) || queue ).to_sym
          config_file_queue(queue).push fileinfo
        else
         config_file_queue(queue).push(:path => :system, :file => fileinfo.to_s)
      end
    end

    #
    #
    #
    def already_configured?
      @config[:core] && @config[:core][:configured]
    end

    def already_configured!
      core[:configured] = true
    end

    def add_section(name)
      @sections << name
      @config[name] ||= {}
      self.instance_eval %{
        def #{name.to_s}
          @config[:#{name.to_s}]
        end
      }
    end

    def add(section, hash)
      add_section section
      apply_configuration({ section => hash })
      @config
    end

    #
    # allow to set specific ovverides
    #
    def override(section, &block)
      @overrides[section] = block
    end

    #
    # this should be called _after_ the additional parse blocks have been defined
    #
    def configure!(options = {})

      return if already_configured?

      # process commandline opts
      parse_options if options.delete(:parse_command_line)

      if(@options && @options[:core])
        @config[:core][:environment] = @options[:core][:environment] if @options[:core][:environment]
        @config[:core][:config_file] = @options[:core][:config_file] if @options[:core][:config_file]
      end

      env = @config[:core][:environment]

      #
      # Tomte.config.add_config_files {
      #   { :file => tomte-protocol.yml, :path => :system }
      #   { :file => ...                 :path => :application
      # }
      #
      config_file_queues.keys.sort.each do |queue|
        config_file_queue(queue).each do |f|
          [@config[:core][:common_environment], env].each do |e|
              apply_configuration load_configuration_from f, e
          end
        end
      end

      # also merge the command line options in
      apply_configuration @options if @options
      # and the local options
      apply_configuration options

      # apply overrides
      apply_overrides

      # a few defaults actually depends on the configuration given
      post_configuration

      already_configured!

      tomte_debug do
        puts "== parsed configuration ==".yellow.bold
        ap @config
      end

      @config
    end

    # quickies
#    def core
#      self[:core]
#    end

#    def local
#      self[:local]
#    end

    def to_s
      @config.to_s
    end

    def debug
      @config.ai.subindent
    end

    def debug?
      !!@config[:core][:debug]
    end

    #
    # register additional options definitions
    #
    # config.additional_options(name) do |cf, opts|
    #
    #   opts.on("-d", "--daemonize", "Run  as a daemon") do |d|
    #      cf[:daemonize] = d
    #    end
    #
    # end
    #
    def additional_options(name, &blk)
      @additional_options_blocks[name] = blk
    end

    def additional_commands(hash)
      (@additional_commands ||= {}).merge! hash
    end

    def additional_help_text(text)
      (@additional_help_text ||= []) << text
    end

    private
    #######
    #
    #
    #

    def apply_overrides
      @sections.each do |s|
        @overrides[s].call(@config[s]) if(@overrides[s] &&  @overrides[s].respond_to?(:call))
      end
    end


    def default_environment
      if !ENV['TOMTE_ENV'].blank?
        core[:environment] = ENV['TOMTE_ENV'].strip
      elsif File.exists?(DEFAULT_ENVIRONMENT_FILE)
        core[:environment] = File.read(DEFAULT_ENVIRONMENT_FILE).match(/^\w+/)[0]
      end

    rescue
      # do nothing
    end

    def default_configuration
      { :core => DEFAULT_CORE_CONFIGURATION, :local => {} }
    end

    def post_configuration

      # set it to a sensible default
      core[:application_root] = File.expand_path('..', Tomte::LIBRARY ) unless core[:application_root]

      local[:uuid] ||= UUIDTools::UUID.random_create.to_s
      local[:host] ||= Socket.gethostname
      local[:identity]  ||= core[:application] + '-' + local[:host]
      begin
        local[:hostname] ||= Socket.gethostbyname(Socket.gethostname).first
      rescue
        local[:hostname] ||= 'unknown'
      end

      # load the debugging extensions
      if debug?
        puts "== Debugging active. This is going to be slow!".important!
        Object.__send__(:include, Tomte::Debug::Active)
      end

    end


    def expand_path_spec(file, path)
      cwd = Dir.getwd

      case path
        when :system
          File.expand_path(file, core[:system_configuration_path])
        when :application
          File.expand_path(file, File.join(core[:application_root], 'config'))
        when :cwd
          File.expand_path(file, cwd)
        when :nil
          File.expand_path(file)
        else
          File.expand_path(file, path.to_s)
        end
    end

    def load_configuration_from(file, environment = 'default', path = nil)

      (file, path) = file[:file], file[:path] if file.is_a? Hash
      file = expand_path_spec(file, path)

      puts "= Loading configuration file #{file}, for environment: #{environment}".blue if @options &&
        @options[:core] && @options[:core][:debug]

      begin
        config = {}
        raw_config = if File.exists?(file) 
          data = File.read file
          tpl = ERB.new data
          yaml = tpl.result binding
          YAML.load(yaml) || {}
        else
          {}
        end

        config = raw_config[environment] if raw_config.has_key?(environment)
        config.recursively!{|h| h.symbolize_keys }
      rescue Exception => e
        puts "Error while reading configuration file '#{file}'".important!
        puts " - " + e.message
        exit
      end
      config
    end



    def parse_options(&blk)

      @argv = core[:argv] || ARGV
      @options = {}
      @sections.each {|s| @options[s] = {}}

      opts = OptionParser.new do |opts|

        opts.banner = usage

        opts.define_head(core[:application_caption].green.bold) if core[:application_caption]

        if @additional_help_text
          opts.separator " "
          @additional_help_text.each do |l|
            opts.separator l.to_s
          end
        end

        unless @additional_commands.blank?
          opts.separator " "
          opts.separator 'Commands'.blue.bold
          opts.separator SEPARATOR
          @additional_commands.each do |command, caption|
            opts.separator sprintf("%-25s", command).yellow.bold + caption
          end
        end

        opts.separator " "
        opts.separator 'Tomte Options'.blue.bold
        opts.separator SEPARATOR

        opts.on("-e", "--environment ENV", "Framework environment " + "(default: #{@config[:core][:environment]})") do |env|
          @options[:core][:environment] = env
        end

        opts.on("-i", "--identity ID", "Sets this endpoint identity") do |ident|
          @options[:local][:identity] = ident
        end


        opts.on("-D", "--debug", "Set debugging on (heavy and verbose!)") do
          @options[:core][:debug] = true
          @options[:core][:log_profile] = :debug
          @options[:core][:log_level] = :debug
        end

        opts.on("-l", "--log-level LEVEL", "Specify the log level (fatal, error, warn, info, debug). Default is '#{@config[:log_level]}'") do |level|
          @options[:core][:log_level] = level
        end

        opts.on("--log-path PATH", "Specify the log file path") do |dir|
          @options[:core][:log_profile] = :production
          @options[:core][:log_path] = dir
        end

        opts.on("--log-profile LEVEL", "Specify the log profile to use [interactive | production | daemonized | ...]") do |profile|
          @options[:core][:log_profile] = profile
        end
        opts.on("-h", "--help", "Show this message")                               { puts opts; exit }

        opts.on("--version", "Show the Tomte version number") do |res|
          puts "Tomte Version #{Tomte::VERSION}"
          exit
        end

         # call additional blocks, usually additional options in the agent(s)
        blk.call(@config, opts) if blk
        @additional_options_blocks.each do |name, blk|
          opts.separator " "
          opts.separator name.blue.bold
          opts.separator SEPARATOR
          blk.call(@options, opts, @config)
        end

      end

      opts.parse! @argv
      @options[:core][:command] = @argv.shift if @argv
      @options[:core][:arguments] = @argv

      if (@additional_commands && @options[:core][:command] && !@additional_commands.keys.include?(@options[:core][:command].to_sym))
        puts "\nUnknown command '#{@options[:core][:command]}'".important!
        puts usage
        exit
      end
    end

    def usage
      ("Usage: #{core[:application]} [-flags] [--options] " + "#{output_commands}\n").blue.bold
    end

    def output_commands
      (!@additional_commands.blank? ? "[#{@additional_commands.keys.join('|')}] [params]" : '')
    end

    class << self
      def relative_cfg(library_path)
        File.expand_path(File.join(library_path, '..', 'config'))
      end
    end

  end
end
