# encoding: UTF-8
#
# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2013, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

class GenericError < StandardError

  attr_reader :type
  attr_reader :attributes
end


module Tomte
  module Document
    module Parser

      class << self
        def from_bson(data)
          load(BSON.deserialize(data))
        end

        def is_document?(attrs)
          return false unless attrs.kind_of? Hash
          attrs['_doctype'].present? || attrs[:_doctype].present?
        end

        def is_generic_object?(attrs)
          return false unless attrs.kind_of? Hash
          attrs['_type'].present? || attrs[:_type].present?
        end

        def load_generic_object(attrs)
          if is_generic_object? attrs
            (attrs['_type'] || attrs[:_type]).constantize.from_tomte(attrs)
          else
            attrs
          end
        rescue NameError
          OpenStruct.new(attrs)
        end

        def parseable?(attrs)
          is_document?(attrs) || is_generic_object?(attrs) || attrs.is_a?(::Hash)
        end

        def load_document(attrs)
          if is_document? attrs
            (attrs['_doctype'] || attrs[:_doctype]).constantize.from_tomte(attrs)
          else
            attrs
          end

        rescue NameError
          Tomte::OpenDocument.new(attrs)
        end

        #
        # returns an exception with the given message, backtrace and attributes hash
        #
        def load_exception(klass, message = '', backtrace = [], attributes = {})

          e = begin
                klass.constantize
              rescue NameError
                GenericError
              end
          exc =
            begin
              raise(e.allocate, message)
            rescue e => exception
              exception.set_backtrace(backtrace)
              exception
            end

          if exc.is_a?(GenericError)
            exc.instance_variable_set(:@type, klass) if exc.is_a?(GenericError)
            exc.instance_variable_set(:@attributes, OpenStruct.new(attributes.inject_hash { |h,(k,v)| h[k] = load(v) }))
          else
            attributes.each do |k,v|
              exc.instance_variable_set("@#{k}".to_sym, load(v))
            end
          end
          exc
        end

        def load(attrs)
          if is_document? attrs
            load_document(attrs)
          elsif is_generic_object? attrs
            load_generic_object(attrs)
          elsif attrs.is_a? ::Hash
            ActiveSupport::HashWithIndifferentAccess.from_tomte(attrs)
          elsif attrs.is_a? ::Array
            Array.from_tomte(attrs)
          else
            attrs
          end
        end
        alias_method :from_tomte, :load

        def to_tomte_hash(obj)
          return(obj) unless obj.respond_to?(:to_tomte)

          klass = obj.class
          res = obj.to_tomte

          if res.is_a?(::Hash) &&
                       !Tomte::Document::Parser.is_document?(res) &&
                       !Tomte::Document::Parser.is_generic_object?(res) &&
                       !klass.ancestors.include?(::Hash)

            res[:_type] = klass.name
          end
          res
        end
      end
    end
  end
end